package sun.reflect.annotation;

import java.lang.annotation.AnnotationFormatError;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.List;
import java.lang.reflect.AnnotatedElement;
import java.lang.annotation.Annotation;

public final class TypeAnnotation
{
    private final TypeAnnotation.TypeAnnotationTargetInfo targetInfo;
    private final TypeAnnotation.LocationInfo loc;
    private final Annotation annotation;
    private final AnnotatedElement baseDeclaration;
    
    public TypeAnnotation(TypeAnnotation.TypeAnnotationTargetInfo targetInfo, TypeAnnotation.LocationInfo loc, Annotation annotation, AnnotatedElement baseDeclaration) {
        this.targetInfo = targetInfo;
        this.loc = loc;
        this.annotation = annotation;
        this.baseDeclaration = baseDeclaration;
    }
    
    public TypeAnnotation.TypeAnnotationTargetInfo getTargetInfo() {
        return this.targetInfo;
    }
    
    public Annotation getAnnotation() {
        return this.annotation;
    }
    
    public AnnotatedElement getBaseDeclaration() {
        return this.baseDeclaration;
    }
    
    public TypeAnnotation.LocationInfo getLocationInfo() {
        return this.loc;
    }
    
    public static List<TypeAnnotation> filter(TypeAnnotation[] array, TypeAnnotation.TypeAnnotationTarget typeAnnotationTarget) {
        ArrayList list = new ArrayList(array.length);
        for (TypeAnnotation e : array) {
            if (e.getTargetInfo().getTarget() == typeAnnotationTarget) {
                list.add(e);
            }
        }
        list.trimToSize();
        return list;
    }
    
    @Override
    public String toString() {
        return this.annotation.toString() + " with Targetnfo: " + this.targetInfo.toString() + " on base declaration: " + this.baseDeclaration.toString();
    }
    
    public enum TypeAnnotationTarget
    {
        CLASS_TYPE_PARAMETER, 
        METHOD_TYPE_PARAMETER, 
        CLASS_EXTENDS, 
        CLASS_IMPLEMENTS, 
        CLASS_TYPE_PARAMETER_BOUND, 
        METHOD_TYPE_PARAMETER_BOUND, 
        FIELD, 
        METHOD_RETURN, 
        METHOD_RECEIVER, 
        METHOD_FORMAL_PARAMETER, 
        THROWS;
    }
    
    public static final class TypeAnnotationTargetInfo
    {
        private final TypeAnnotation.TypeAnnotationTarget target;
        private final int count;
        private final int secondaryIndex;
        private static final int UNUSED_INDEX = -2;
        
        public TypeAnnotationTargetInfo(TypeAnnotation.TypeAnnotationTarget typeAnnotationTarget) {
            this(typeAnnotationTarget, -2, -2);
        }
        
        public TypeAnnotationTargetInfo(TypeAnnotation.TypeAnnotationTarget typeAnnotationTarget, int n) {
            this(typeAnnotationTarget, n, -2);
        }
        
        public TypeAnnotationTargetInfo(TypeAnnotation.TypeAnnotationTarget target, int count, int secondaryIndex) {
            this.target = target;
            this.count = count;
            this.secondaryIndex = secondaryIndex;
        }
        
        public TypeAnnotation.TypeAnnotationTarget getTarget() {
            return this.target;
        }
        
        public int getCount() {
            return this.count;
        }
        
        public int getSecondaryIndex() {
            return this.secondaryIndex;
        }
        
        @Override
        public String toString() {
            return "" + this.target + ": " + this.count + ", " + this.secondaryIndex;
        }
    }
    
    public static final class LocationInfo
    {
        private final int depth;
        private final TypeAnnotation.LocationInfo.Location[] locations;
        public static final TypeAnnotation.LocationInfo BASE_LOCATION;
        
        private LocationInfo() {
            this(0, new TypeAnnotation.LocationInfo.Location[0]);
        }
        
        private LocationInfo(int depth, TypeAnnotation.LocationInfo.Location[] locations) {
            this.depth = depth;
            this.locations = locations;
        }
        
        public static TypeAnnotation.LocationInfo parseLocationInfo(ByteBuffer byteBuffer) {
            int n = byteBuffer.get() & 0xFF;
            if (n == 0) {
                return TypeAnnotation.LocationInfo.BASE_LOCATION;
            }
            TypeAnnotation.LocationInfo.Location[] array = new TypeAnnotation.LocationInfo.Location[n];
            int i = 0;
            while (i < n) {
                int value = byteBuffer.get();
                int n2 = (short)(byteBuffer.get() & 0xFF);
                if (value != 0 && !(value == 1 | value == 2) && value != 3) {
                    throw new AnnotationFormatError("Bad Location encoding in Type Annotation");
                }
                if (value != 3 && n2 != 0) {
                    throw new AnnotationFormatError("Bad Location encoding in Type Annotation");
                }
                array[i] = new TypeAnnotation.LocationInfo.Location((byte)value, (short)n2);
                ++i;
            }
            return new TypeAnnotation.LocationInfo(n, array);
        }
        
        public TypeAnnotation.LocationInfo pushArray() {
            return this.pushLocation((byte)0, (short)0);
        }
        
        public TypeAnnotation.LocationInfo pushInner() {
            return this.pushLocation((byte)1, (short)0);
        }
        
        public TypeAnnotation.LocationInfo pushWildcard() {
            return this.pushLocation((byte)2, (short)0);
        }
        
        public TypeAnnotation.LocationInfo pushTypeArg(short n) {
            return this.pushLocation((byte)3, n);
        }
        
        public TypeAnnotation.LocationInfo pushLocation(byte b, short n) {
            int n2 = this.depth + 1;
            TypeAnnotation.LocationInfo.Location[] array = new TypeAnnotation.LocationInfo.Location[n2];
            System.arraycopy(this.locations, 0, array, 0, this.depth);
            array[n2 - 1] = new TypeAnnotation.LocationInfo.Location(b, (short)(n & 0xFF));
            return new TypeAnnotation.LocationInfo(n2, array);
        }
        
        public TypeAnnotation[] filter(TypeAnnotation[] array) {
            ArrayList list = new ArrayList(array.length);
            for (TypeAnnotation e : array) {
                if (this.isSameLocationInfo(e.getLocationInfo())) {
                    list.add(e);
                }
            }
            return (TypeAnnotation[])list.toArray(new TypeAnnotation[0]);
        }
        
        boolean isSameLocationInfo(TypeAnnotation.LocationInfo locationInfo) {
            if (this.depth != locationInfo.depth) {
                return false;
            }
            int i = 0;
            while (i < this.depth) {
                if (!this.locations[i].isSameLocation(locationInfo.locations[i])) {
                    return false;
                }
                ++i;
            }
            return true;
        }
        
        static {
            BASE_LOCATION = new TypeAnnotation.LocationInfo();
        }
        
        public static final class Location
        {
            public final byte tag;
            public final short index;
            
            boolean isSameLocation(TypeAnnotation.LocationInfo.Location location) {
                return this.tag == location.tag && this.index == location.index;
            }
            
            public Location(byte tag, short index) {
                this.tag = tag;
                this.index = index;
            }
        }
    }
}
